/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include "threadlog/logc.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";

/*
 * Private Functions
 */

static void next_logid(struct logc *logc)
{
	do {
		logc->seq++;
	} while (logc->seq == LOGID_NUL || logc->seq == LOGID_NONE);
}

static int connect_all(struct logc *logc, char const *server_name, unsigned short port)
{
	logc->sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == logc->sockfd) {
		fprintf(stderr, "Cannot create socket : %s\n", strerror(errno));
		return -1;
	}
	struct hostent *hptr;
	if (NULL == (hptr = gethostbyname(server_name))) {
		fprintf(stderr, "Cannot resolve host %s : %s\n", server_name, hstrerror(h_errno));
		return -1;
	}
	if (hptr->h_addrtype != AF_INET) {
		fprintf(stderr, "Unknown address type\n");
		return -1;
	}
	char **addr = hptr->h_addr_list;
	if (! *addr) {
		fprintf(stderr, "No address found for %s\n", server_name);
		return -1;
	}
	memset(&logc->server, 0, sizeof(logc->server));
	logc->server.sin_family = AF_INET;
	logc->server.sin_port = htons(port);
	logc->server.sin_addr = *(struct in_addr *) *addr;
	return 0;
}

static logid_t my_print(struct log *log, enum log_level level, time_t date, logid_t *refs, char const *format, ...)
{
	assert(log && format);
	struct logc *logc = (void *)((char *)log - offsetof(struct logc, log));
	if (level > logc->log.level) return LOGID_NONE;
	next_logid(logc);
	struct dtg_max dtg = {
		.hdr = {
			.level = level,
			.date = date,
			.this_logid = logc->seq,
		},
	};
	memcpy(&dtg.hdr.local_id, logc->local_id, sizeof(*logc->local_id));
	va_list ap;
	va_start(ap, format);
	dtg.hdr.str_size = vsnprintf(dtg.string, sizeof(dtg.string), format, ap);
	va_end(ap);
	if (dtg.hdr.str_size < 0) {
		fprintf(stderr, "Cannot send this string (format = '%s')\n", format);
		return LOGID_NONE;
	}
	dtg.hdr.str_size++;		// we want trailing NUL
	unsigned r = 0;
	if (refs) {
		for (; r < sizeof_array(dtg.hdr.refs) && refs[r] != LOGID_NUL; r++) {
			dtg.hdr.refs[r] = refs[r];
		}
	}
	if (r < sizeof_array(dtg.hdr.refs)) {
		dtg.hdr.refs[r] = LOGID_NUL;
	}
	// send this message (DONTWAIT should not be necessary as this is DGRAM)
	ssize_t msg_size = sizeof(struct dtg_hdr) + dtg.hdr.str_size;
	ssize_t ret =
		sendto(logc->sockfd, &dtg, msg_size, MSG_DONTWAIT, (struct sockaddr *)&logc->server, sizeof(logc->server));
	if (ret < 0) {
		fprintf(stderr, "Cannot send log '%s' : %s\n", dtg.string, strerror(errno));
		return LOGID_NONE;
	}
	if (ret != msg_size) {
		fprintf(stderr, "Message '%s' was truncated\n", dtg.string);
		// but we go on
	}
	return logc->seq;
}

/*
 * Public Functions
 */

int logc_ctor(struct logc *logc, char (*local_id)[8], char const *server_name, unsigned short port)
{
	assert(logc && local_id && server_name);
	if (0 != log_ctor(&logc->log)) {
		return -1;
	}
	static struct log_ops const my_ops = {
		.print = my_print,
	};
	logc->local_id = local_id;
	logc->log.ops = &my_ops;
	logc->seq = 0;		// anything would do
	if (0 != connect_all(logc, server_name, port)) {
		return -1;
	}
	return 0;
}

void logc_dtor(struct logc *logc)
{
	if (-1 != logc->sockfd) {
		(void)close(logc->sockfd);
		logc->sockfd = -1;
	}
}

