/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>		// We log onto stderr
#include <assert.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "threadlog/logd.h"
#include "threadlog/logdlib.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";

/*
 * Private Functions
 */

static int rotate_file(struct logd *logd, logid_t next_id)
{
	int msgs_fd = get_fd_from_first_id(next_id, logd->path);
	if (msgs_fd == -1) {
		return -1;
	}
	logd->msgs_fd = msgs_fd;
	logd->first_id = next_id;
	return 0;
}

static int append_msg(struct logd *logd, void const *str, size_t size)
{
	return write_all(logd->msgs_fd, str, size);
}

static int back_reference(struct logd *logd, logid_t refered, logid_t referent)
{
	int fd = open_backlink(refered, logd->path);
	if (-1 == fd) {
		return -1;
	}
	int ret = write_all(fd, &referent, sizeof(referent));
	(void)close_backlink(refered, fd);
	return ret;
}

static void close_all(struct logd *logd)
{
	if (-1 != logd->msgs_fd) {
		(void)close(logd->msgs_fd);
	}
}

static logid_t my_print(struct log *log, enum log_level level, time_t date, logid_t *refs_, char const *format, ...)
{
	assert(log && format);
	struct logd *logd = (void *)((char *)log - offsetof(struct logd, log));
	if (level > logd->log.level) return LOGID_NONE;
	size_t fsize = file_size(logd->msgs_fd);
	logid_t id = logd->first_id + fsize;
	if (logid_reserved(id) || fsize > logd->max_filesize) {
		while (logid_reserved(id)) id++;
		rotate_file(logd, id);
	}
	struct msg msg;
	va_list ap;
	va_start(ap, format);
	msg.record.str_size = vsnprintf(msg.string, sizeof(msg.string), format, ap);
	va_end(ap);
	if (msg.record.str_size < 0) {
		fprintf(stderr, "Cannot print this string (format = '%s')\n", format);
		return LOGID_NONE;
	}
	msg.record.str_size ++;	// we want trailing NUL
	if ((size_t)msg.record.str_size > sizeof(msg.string)) {
		fprintf(stderr, "String was truncated by %d chars\n",
			msg.record.str_size - sizeof(msg.string));
		msg.record.str_size = sizeof(msg.string);
	}
	if (msg.record.str_size & 3) {	// align on 4 bytes for faster reading later
		msg.record.str_size = (msg.record.str_size & ~3) + 4;
	}
	// Build the record
	msg.record.date = date;
	msg.record.level = level;
	unsigned r = 0;
	logid_t *refs = refs_;
	if (! refs) {
		refs = logd->default_refs;
	}
	for (; r < sizeof_array(msg.record.refs) && refs[r] != LOGID_NUL; r++) {
		if (refs[r] < logd->default_refs[0] || refs[r] >= id) {
			fprintf(stderr, "Bad ref : %lu\n", (unsigned long) refs[r]);
			return LOGID_NONE;
		}
		msg.record.refs[r] = refs[r];
	}
	for (; r < sizeof_array(msg.record.refs); r++) {
		msg.record.refs[r] = LOGID_NUL;
	}
	if (0 != append_msg(logd, &msg, sizeof(msg.record) + msg.record.str_size)) {
		return LOGID_NONE;
	}
	// Now try to keep back refs.
	for (
		r = 0;
		r < sizeof_array(msg.record.refs) && msg.record.refs[r] != LOGID_NUL;
		r++
	) {
		if (msg.record.refs[r] != LOGID_NONE) {
			(void)back_reference(logd, msg.record.refs[r], id);
		}
	}
	return id;
}

/*
 * Public Functions
 */

int logd_ctor(struct logd *logd, char const *log_path, size_t max_filesize_mb)
{
	assert(logd && log_path);
	if (0 != log_ctor(&logd->log)) {
		return -1;
	}
	static struct log_ops const my_ops = {
		.print = my_print,
	};
	logd->log.ops = &my_ops;
	logd->path = log_path;
	logd->max_filesize = max_filesize_mb * 1024 * 1024;
	if (0 != mkdir_all(logd->path)) {
		return -1;
	}
	if (0 != get_last_fileid(&logd->first_id, logd->path)) {
		return -1;
	}
	logd->msgs_fd = get_fd_from_first_id(logd->first_id, logd->path);
	if (-1 == logd->msgs_fd) {
		return -1;
	}
	if (0 != trunc_to_min_size(logd->msgs_fd, sizeof(struct log_record) + 1)) {
		return -1;
	}
	logd->default_refs[0] = LOGID_NUL;
	logd->default_refs[1] = LOGID_NUL;
	// if the file is empty, log a file creation msg.
	if (file_size(logd->msgs_fd) == 0) {
		logd->default_refs[0] = my_print(&logd->log, log_ALWAYS, time(NULL), NULL,
			"Creating new file with first id #%u at %s",
			(unsigned) logd->first_id, logd->path);
		if (logd->default_refs[0] == LOGID_NONE) {
			close_all(logd);
			return -1;
		}
	} else {
		logd->default_refs[0] = logd->first_id;
	}
	// log a 'welcome' message referencing the first from this file, and use it as initial reference
	logd->default_refs[0] =
		my_print(&logd->log, log_ALWAYS, time(NULL), logd->default_refs,
			"Starting logd in %s", logd->path);
	if (logd->default_refs[0] == LOGID_NONE) {
		close_all(logd);
		return -1;
	}
	return 0;
}

void logd_dtor(struct logd *logd)
{
	my_print(&logd->log, log_NOTICE, time(NULL), logd->default_refs, "Stoping logd");
	close_all(logd);
}

