/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
/* Everything thats usefull to manipulate a logd directory,
 * used both by logd lib and delog tool.
 */
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>		// We log onto stderr
#include <assert.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "threadlog/logd.h"
#include "threadlog/logdlib.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";

#define MSGS_FILENAME_PREFIX "msgs_"
#define MSGS_FILENAME_LEN 5
#define BACKLINKS_FILENAME_PREFIX "bl_"
#define BACKLINKS_FILENAME_LEN 3

/*
 * Private Functions
 */

static int msgs_filter(const struct dirent *dentry)
{
	if (0 != strncmp(dentry->d_name, MSGS_FILENAME_PREFIX, MSGS_FILENAME_LEN)) {
		return 0;
	}
	// Check that the file countain at least a record and an empty msg
	return filename_size(dentry->d_name) > (ssize_t) sizeof(struct log_record);
}

static int msgs_compar(void const *a_, void const *b_)
{
	struct dirent const *const *a = a_;
	struct dirent const *const *b = b_;
	// We sort by decreasing first_date order.
	time_t time_a, time_b;
	if (0 != read_all_filename(&time_a, (*a)->d_name, sizeof(time_a))) {
		return 0;		// ??
	}
	if (0 != read_all_filename(&time_b, (*b)->d_name, sizeof(time_b))) {
		return 0;		// ??
	}
	return time_a > time_b ? -1 : 1;
}

/*
 * Public Functions
 */

int mkdir_all(char const *path)
{
	char filename[PATH_MAX];
	snprintf(filename, sizeof(filename), "%s", path);
	char *c = filename;
	if (! *c) {
		fprintf(stderr, "Empty string is not a wise path");
		return -1;
	}
	for (c = filename; *(c++);) {
		if ('/' == *c || '\0' == *c) {
			char tmp = *c;
			*c = '\0';
			if (-1 == mkdir(filename, 0744) && EEXIST != errno) {
				fprintf(stderr, "mkdir '%s' : %s\n", filename, strerror(errno));
				return -1;
			}
			*c = tmp;
		}
	}
	return 0;
}

int write_all(int fd, void const *buf, size_t size)
{
	size_t written = 0;
	while (written < size) {
		ssize_t ret = write(fd, (char *) buf + written, size - written);
		if (-1 == ret) {
			fprintf(stderr, "Cannot write to file %d : %s\n", fd, strerror(errno));
			return -1;
		}
		written += ret;
	}
	return 0;
}

int read_all_or_EOF(int fd, void *buf, size_t size)
{
	size_t reads = 0;
	while (reads < size) {
		ssize_t ret = read(fd, (char *) buf + reads, size - reads);
		if (-1 == ret) {
			fprintf(stderr, "Cannot read from file %d : %s\n", fd, strerror(errno));
			return -1;
		}
		if (0 == ret) {
			if (0 == reads) return 0;
			fprintf(stderr, "Short read from file %d\n", fd);
			return -1;
		}
		reads += ret;
	}
	return 1;
}

int read_all(int fd, void *buf, size_t size)
{
	switch (read_all_or_EOF(fd, buf, size)) {
		default:
			return -1;
		case 0:
			fprintf(stderr, "Short read from file %d\n", fd);
			return -1;
		case 1:
			return 0;
	}
}

ssize_t file_size(int fd)
{
	struct stat bufstat;
	if (0 != fstat(fd, &bufstat)) {
		fprintf(stderr, "Cannot stat file (%d) : %s\n", fd, strerror(errno));
		return -1;
	}
	return bufstat.st_size;
}

ssize_t filename_size(char const *file)
{
	struct stat bufstat;
	if (0 != stat(file, &bufstat)) {
		fprintf(stderr, "Cannot stat file %s : %s\n", file, strerror(errno));
		return -1;
	}
	return bufstat.st_size;
}

int trunc_to_min_size(int fd, size_t min_size)
{
	ssize_t fsize = file_size(fd);
	if (fsize == -1) {
		return -1;
	}
	if (fsize > 0 && (size_t) fsize < min_size) {
		if (0 != ftruncate(fd, 0)) {
			fprintf(stderr, "Cannot truncate short file (%d) : %s\n", fd, strerror(errno));
			return -1;
		}
	}
	return 0;
}

int read_all_filename(void *buf, char const *filename, size_t size)
{
	int fd = open(filename, O_RDONLY);
	if (-1 == fd) {
		fprintf(stderr, "Cannot open for read_all %s : %s\n", filename, strerror(errno));
		return -1;
	}
	int ret = read_all(fd, buf, size);
	(void)close(fd);
	return ret;
}

int get_namelist(struct dirent ***namelist, char const *path)
{
	// For scandir to work, we need to be in the log_path directory
	char *pwd = getcwd(NULL, 0);
	if (! pwd) {
		fprintf(stderr, "Cannot getcwd : %s\n", strerror(errno));
		return -1;
	}
	if (0 != chdir(path)) {
		fprintf(stderr, "Cannot chdir to log path %s : %s\n", path, strerror(errno));
		free(pwd);
		return -1;
	}
	// find the last logid, using the date stored in the ref file
	int nb_dentry = scandir(path, namelist, msgs_filter, msgs_compar);
	if (-1 == nb_dentry) {
		fprintf(stderr, "Cannot scandir %s : %s\n", path, strerror(errno));
	} else if (0 == nb_dentry) {
		free(*namelist);	// assume that even in this case namelist was mallocated
	}
	if (0 != chdir(pwd)) {
		fprintf(stderr, "Cannot go back to PWD %s : %s\n", pwd, strerror(errno));
		free(*namelist);
		nb_dentry = -1;
	}
	free(pwd);
	return nb_dentry;
}

int get_last_fileid(logid_t * id, char const *path)
{
	struct dirent **namelist;
	int nb_dentry = get_namelist(&namelist, path);
	if (-1 == nb_dentry) {
		return -1;
	}
	if (0 == nb_dentry) {
		*id = 0;		// will start from first id
		while (*id == LOGID_NONE || *id == LOGID_NUL) (*id)++;
		return 0;
	}
	char *end;
	unsigned long ul = strtoul(namelist[0]->d_name + MSGS_FILENAME_LEN, &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "Bad msgs filename : %s\n", namelist[0]->d_name);
		free(namelist);
		return -1;
	}
	*id = ul;
	free(namelist);
	return 0;
}

logid_t get_first_id(logid_t id, char const *path)
{
	struct dirent **namelist;
	int nb_dentry = get_namelist(&namelist, path);
	if (nb_dentry <= 0) {
		return -1;
	}
	char *end;
	logid_t first = LOGID_NONE;
	for (int e = 0; e < nb_dentry; e++) {
		unsigned long ul = strtoul(namelist[e]->d_name + MSGS_FILENAME_LEN, &end, 10);
		if (*end != '\0') {
			fprintf(stderr, "Bad msgs filename : %s\n", namelist[0]->d_name);
			continue;
		}
		if (ul > id) break;
		first = ul;
	}
	free(namelist);
	return first;
}

int get_fd_from_first_id(logid_t id, char const *path)
{
	char filename[PATH_MAX];
	int len = snprintf(filename, sizeof(filename),
		"%s/%s%lu", path, MSGS_FILENAME_PREFIX, (unsigned long) id);
	if (len < 0 || (size_t) len >= sizeof(filename)) {
		fprintf(stderr, "Filename too long : %s\n", filename);
		return -1;
	}
	int fd = open(filename, O_RDWR | O_CREAT | O_APPEND, 0600);
	if (-1 == fd) {
		fprintf(stderr, "Cannot open %s : %s\n", filename, strerror(errno));
	}
	return fd;
}

static int get_fd_from_id(logid_t id, char const *path)
{
	logid_t first = get_first_id(id, path);
	if (LOGID_NONE == first) {
		return -1;
	}
	int fd = get_fd_from_first_id(first, path);
	if (-1 == fd) {
		return -1;
	}
	if ((off_t)-1 == lseek(fd, id - first, SEEK_SET)) {
		fprintf(stderr, "Cannot read_msg (lseek) : %s\n", strerror(errno));
		close(fd);
		return -1;
	}
	return fd;
}

int read_msg(struct log_record *record, logid_t id, char const *path)
{
	int err = -1;
	int fd = get_fd_from_id(id, path);
	if (-1 == fd) {
		goto rm_0;
	}
	if (0 != read_all(fd, record, sizeof(*record))) {
		goto rm_1;
	}
	if (0 != read_all(fd, record->string, record->str_size)) {
		goto rm_1;
	}
	err = 0;
rm_1:
	close(fd);
rm_0:
	return err;
}

int get_backlink_name(char *str, char const *path, size_t max_size, logid_t id)
{
	int len =
		snprintf(str, max_size, "%s/" BACKLINKS_FILENAME_PREFIX "%lu", path, (unsigned long) id);
	if (len < 0 || (size_t) len >= max_size) {
		fprintf(stderr, "Backlink filename too long ; %s\n", str);
		return -1;
	}
	return 0;
}

int open_backlink(logid_t id, char const *path)
{
	// TODO: use a cache for filedescrs
	char filename[PATH_MAX];
	if (0 != get_backlink_name(filename, path, sizeof(filename), id)) {
		return -1;
	}
	int fd = open(filename, O_RDWR | O_CREAT | O_APPEND, 0600);
	if (-1 == fd) {
		fprintf(stderr, "Cannot open %s : %s\n", filename, strerror(errno));
	}
	return fd;
}

int close_backlink(logid_t id, int fd)
{
	// TODO: use a cache for filedescrs
	(void)id;
	return close(fd);
}

unsigned nb_bl(logid_t id, char const *path)
{
	char filename[PATH_MAX];
	if (0 != get_backlink_name(filename, path, sizeof(filename), id)) {
		return 0;
	}
	struct stat buf;		// cannot use filename_size because file may not exist
	if (0 == stat(filename, &buf)) {
		return buf.st_size / sizeof(logid_t);
	} else {
		if (ENOENT != errno) {
			fprintf(stderr, "Cannot stat file '%s' : %s\n", filename, strerror(errno));
		}
		return 0;
	}
}

// TODO: implements with a mmap and return a pointer to the record & string ?
int log_msg_ctor(struct log_msg *msg, logid_t id, char const *path)
{
	assert(path && msg);
	if (0 != read_msg(&msg->record, id, path)) {
		return -1;
	}
	msg->id = id;
	msg->nb_bl = nb_bl(id, path);
	return 0;
}

void log_msg_dtor(struct log_msg *msg)
{
	(void)msg;
}

logid_t path_first_id(char const *path)
{
	struct dirent **namelist;
	int nb_dentry = get_namelist(&namelist, path);
	if (nb_dentry <= 0) {
		return LOGID_NONE;
	}
	char *end;
	unsigned long ul =
		strtoul(namelist[nb_dentry - 1]->d_name + MSGS_FILENAME_LEN, &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "Bad msgs filename : %s\n", namelist[nb_dentry]->d_name);
		free(namelist);
		return LOGID_NONE;
	}
	free(namelist);
	return ul;
}

int bl_curs_ctor(struct bl_curs *curs, logid_t id, char const *path)
{
	assert(curs && path);
	curs->id = id;
	curs->fd = open_backlink(id, path);
	if (curs->fd == -1)
		return -1;
	return 0;
}

void bl_curs_dtor(struct bl_curs *curs)
{
	assert(curs);
	(void)close_backlink(curs->id, curs->fd);
}

logid_t bl_curs_next(struct bl_curs *curs)
{
	assert(curs);
	logid_t id;
	switch (read_all_or_EOF(curs->fd, &id, sizeof(id))) {
		case 0:
			return LOGID_NUL;
		case 1:
			return id;
		default:
			return LOGID_NONE;
	}
}

int chrono_curs_ctor(struct chrono_curs *curs, logid_t id, char const *path)
{
	assert(curs && path);
	curs->first = get_first_id(id, path);
	if (LOGID_NONE == curs->first) {
		return -1;
	}
	curs->fd = get_fd_from_first_id(curs->first, path);
	if (-1 == curs->fd) {
		return -1;
	}
	return 0;
}

void chrono_curs_dtor(struct chrono_curs *curs)
{
	assert(curs);
	(void)close(curs->fd);
}

logid_t chrono_curs_next(struct chrono_curs *curs)
{
	assert(curs);
	struct msg msg;
	off_t pos = lseek(curs->fd, 0, SEEK_CUR);
	switch (read_all_or_EOF(curs->fd, &msg.record, sizeof(msg.record))) {
		case 0:
			// TODO: next file...
			return LOGID_NUL;
		case 1:
			break;
		default:
			return LOGID_NONE;
	}
	if (0 != read_all(curs->fd, msg.record.string, msg.record.str_size)) {
		return LOGID_NONE;
	}
	return pos + curs->first;
}

enum log_level log_prio2level(int prio)
{
	if (prio & (LOG_EMERG|LOG_ALERT|LOG_ERR)) return log_ERROR;
	else if (prio & LOG_WARNING) return log_WARNING;
	else if (prio & (LOG_NOTICE|LOG_INFO)) return log_NOTICE;
	else return log_DEBUG;
}


