/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <stddef.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "threadlog/pool.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";

#define ALIGN(x) ( x & 3U ? (x & ~3U) + 4 : x )

// All folowing macros only works once the pool pointed by variable 'pool' is initialized.
// We cannot use a dequeue_t because we need numerical ids instead of pointers
#define OBJ_T struct { \
	char data[pool->storesize]; \
	unsigned prev, next; \
	time_t age; \
	uint32_t seq; \
}
#define GET_OBJ(x) ((OBJ_T *)get_obj(pool, x))
// UNLINK_ID init list head or flag entry as unlinked 
#define UNLINK_ID(x) do { GET_OBJ(x)->next = GET_OBJ(x)->prev = x; } while ( 0 )
#define DATA2OBJ(data) ( void * )( ( char * )data - offsetof( OBJ_T, data ) )

/*
 * Private Functions
 */

// return a pointer to OBJ_T
static inline void *get_obj(struct pool const *pool, unsigned x)
{
	assert(pool);
	assert(x < pool->nb_objs);
	return (void *)(pool->objs + x*sizeof(OBJ_T));
}

static inline unsigned pool_id(struct pool const *pool, void *obj)
{
	size_t x = ((char *)obj - (char *)pool->objs);
	return x / sizeof(OBJ_T);
}

static bool pool_is_valid(struct pool const *pool, void *data)
{
	OBJ_T *obj = DATA2OBJ(data);
	ssize_t x = ((char *)obj - pool->objs);
	return x >= 0 && x/sizeof(OBJ_T) < pool->nb_objs && x%sizeof(OBJ_T) == 0;
}

static int check_list(struct pool const *pool, unsigned first_id)
{
	unsigned entry_id = first_id;
	do {
		unsigned next_entry_id = GET_OBJ(entry_id)->next;
		if (next_entry_id >= pool->nb_objs) {
			pool->log_cb(LOG_ERR, "Next entry out of pool (id %u)", entry_id);
			return -1;
		}
		if (GET_OBJ(next_entry_id)->prev != entry_id) {
			pool->log_cb(LOG_ERR, "Bad bad link (id %u)", next_entry_id);
			return -1;
		}
		entry_id = next_entry_id;
	} while (entry_id != first_id);
	return 0;
}

static void remove_from_pool_list(struct pool *pool, unsigned id)
{
	OBJ_T *obj = get_obj(pool, id);
	GET_OBJ(obj->next)->prev = obj->prev;
	GET_OBJ(obj->prev)->next = obj->next;
}

static void add_to_pool_list(struct pool *pool, unsigned head_id, unsigned id)
{
	OBJ_T *obj = get_obj(pool, id);
	obj->next = GET_OBJ(head_id)->next;
	obj->prev = head_id;
	GET_OBJ(obj->next)->prev = id;
	GET_OBJ(obj->prev)->next = id;
}

/*
 * Public Functions
 */

int pool_ctor(struct pool *pool, char const *pool_file, unsigned nb_objs, size_t objsize, pool_log_func log_cb)
{
	assert(pool && pool_file);
	pool->nb_objs = nb_objs + 2;	// we keep obj 0 for free list head and obj 1 for non-free list
	pool->objsize = objsize;
	pool->storesize = ALIGN(objsize);
	pool->objs_size = sizeof(OBJ_T) * pool->nb_objs;
	pool->objs = NULL;
	pool->log_cb = log_cb ? log_cb : syslog;
	// open the mmaped file, create/extend/trunc it if necessary
	while (-1 == (pool->fd = open(pool_file, O_RDWR))) {
		if (ENOENT == errno) {	// create it
			if (-1 != (pool->fd = open(pool_file, O_RDWR | O_CREAT | O_EXCL, 0600))) {
				break;
			}
			if (EEXIST != errno) {
				pool->log_cb(LOG_ERR, "Cannot create pool file '%s' : %s", pool_file, strerror(errno));
				return -1;
			}
			// EXIST already ? another process raced against us. retry.
			pool->log_cb(LOG_WARNING, "Another process is racing for the same pool file (%s)", pool_file);
		} else {
			pool->log_cb(LOG_ERR, "Cannot open pool file '%s' : %s", pool_file, strerror(errno));
			return -1;
		}
	}
	// extend / trunc
	off_t file_size = lseek(pool->fd, 0, SEEK_END);
	if ((off_t) - 1 == file_size) {
		pool->log_cb(LOG_ERR, "Cannot get file size : %s", strerror(errno));
		goto close_quit;
	}
	if (0 != ftruncate(pool->fd, pool->objs_size)) {
		pool->log_cb(LOG_ERR, "Cannot truncate pool file : %s", strerror(errno));
		goto close_quit;
	}
	// mmap that file
	pool->objs =
		mmap(NULL, pool->objs_size, PROT_READ | PROT_WRITE, MAP_SHARED, pool->fd, 0);
	if (MAP_FAILED == pool->objs) {
		pool->objs = NULL;
		pool->log_cb(LOG_ERR, "Cannot mmap pool file : %s", strerror(errno));
		goto close_quit;
	}
	// link new entries as free
	unsigned i = pool_id(pool, pool->objs + file_size);
	for (size_t sz = file_size; sz < pool->objs_size; sz += sizeof(OBJ_T), i++) {
		if (i <= 1) {
			GET_OBJ(i)->next = GET_OBJ(i)->prev = i;
		} else {
			add_to_pool_list(pool, 0 /* free list */ , i);
		}
	}
	// check free list
	if (0 != check_list(pool, 0)) {
		pool->log_cb(LOG_ERR, "Bad free list : reseting pool");
		pool_reset(pool);
	}
	if (0 != check_list(pool, 1)) {
		pool->log_cb(LOG_ERR, "Bad alloc list : reseting pool");
		pool_reset(pool);
	}
	return 0;
close_quit:
	close(pool->fd);
	pool->fd = -1;
	return -1;
}

void pool_dtor(struct pool *pool)
{
	assert(pool);
	if (0 != munmap(pool->objs, pool->objs_size)) {
		pool->log_cb(LOG_ERR, "Cannot unmap file : %s", strerror(errno));
	}
	if (0 != close(pool->fd)) {
		pool->log_cb(LOG_ERR, "Cannot close file : %s", strerror(errno));
	}
}

void pool_reset(struct pool *pool)
{
	pool->log_cb(LOG_INFO, "Reseting pool");
	GET_OBJ(0)->next = GET_OBJ(0)->prev = 0;
	GET_OBJ(1)->next = GET_OBJ(1)->prev = 1;
	for (unsigned i = 2; i < pool->nb_objs; i++) {
		add_to_pool_list(pool, 0 /* free list */ , i);
		GET_OBJ(i)->seq = 0;
	}
}

void pool_touch(struct pool *pool, void *data)
{
	OBJ_T *obj = DATA2OBJ(data);
	assert(pool_is_valid(pool, &obj->data));
	time(&obj->age);
}

void *pool_alloc(struct pool *pool, unsigned *id_)
{
	OBJ_T *first = get_obj(pool, 0);
	unsigned id = first->next;
	if (id == 0 /* free list head */ ) {
		pool->log_cb(LOG_ERR, "No more storage in pool");
		return NULL;
	}
	if (id_) {
		*id_ = id;
	}
	remove_from_pool_list(pool, id);
	add_to_pool_list(pool, 1 /* alloc list */ , id);
	pool_touch(pool, &GET_OBJ(id)->data);
	return &GET_OBJ(id)->data;
}

void pool_free(struct pool *pool, void *data)
{
	OBJ_T *obj = DATA2OBJ(data);
	assert(pool_is_valid(pool, &obj->data));
	unsigned id = pool_id(pool, obj);
	remove_from_pool_list(pool, id);
	add_to_pool_list(pool, 0 /* free list */ , id);
	obj->seq++;
}

uint64_t pool_key(struct pool const *pool, void *obj)
{
	unsigned id = pool_id(pool, obj);
	return ((uint64_t) GET_OBJ(id)->seq << 32) | id;
}

void *pool_get(struct pool *pool, unsigned id)
{
	assert(id < pool->nb_objs && id != 0);
	return &GET_OBJ(id)->data;
}

int pool_for_each(struct pool *pool, int (*cb) (struct pool const *, void *, unsigned, void *), void *data)
{
	assert(pool && cb);
	OBJ_T *obj = get_obj(pool, 1);
	unsigned id = obj->next;
	while (id != 1) {
		int ret = cb(pool, obj, id, data);
		if (0 != ret) return ret;
		id = GET_OBJ(id)->next;
	}
	return 0;
}

time_t pool_age(struct pool * pool, void *data)
{
	OBJ_T *obj = DATA2OBJ(data);
	assert(pool_is_valid(pool, &obj->data));
	return obj->age - time(NULL);
}

void pool_dump(struct pool *pool)
{
	assert(pool);
	pool->log_cb(LOG_DEBUG, "Pool dump\n");
	for (unsigned i = 0; i < pool->nb_objs; i++) {
		OBJ_T *obj = get_obj(pool, i);
		pool->log_cb(LOG_DEBUG, "Entry %u, prv:%u nxt:%u (%u)\n",
			i, obj->prev, obj->next, (unsigned) obj->age);
	}
}

extern inline unsigned pool_key2id(uint64_t key);

