/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
/** @file log.h
 * @brief Logs with references
 * @version $Id: log.h,v 1.1.1.1 2006/09/06 16:46:49 ccellier Exp $
 *
 * This is a simple library that allow loging into a dedicated directory,
 * or remotely to a log server.
 * If used locally, two programs can not log into the same directory.
 *
 * The interresting thing is that each log message can reference other log messages,
 * so that the viewer can represent logs as a tree instead of a sequence.
 * For example, when you create a message, log it and keep the log id. When you
 * create a context, log it and keep the log id. When you want to log something
 * about that message or/and that context, reference the beforementioned log ids.
 * Using the viewer, you will then be able to see logs hierarchically per message
 * or per context.
 * 
 * This is just an empty shell. For local loging, use logd. For remote loging, use
 * logc.
 */

#ifndef LOG_H_060823
#define LOG_H_060823

#include <inttypes.h>
#include <time.h>
#include <stdbool.h>
#include <threadlog/commons.h>

#define LOG_STRING_MAX 2048	///< Max size of a log message. Longer messages will be truncated.
#define LOG_REFS_MAX 6		///< Max number of references for a messages.

/** Log level.
 * Messages are logged only when their log level is lower or equal to the
 * current log level.
 * You can still reference messages that were not loged.
 */
enum log_level {
	log_ALWAYS,			///< Always log. Used by the log system itself.
	log_ERROR,			///< Signaling an error condition.
	log_WARNING,		///< Signaling a slight error we can cope with.
	log_NOTICE,			///< Signaling an unusual condition.
	log_DEBUG,			///< Debug messages.
};

typedef uint32_t logid_t;	///< Used as a message id. Use this id to reference the message in others.
#define LOGID_NUL 0		///< Used to end array of references
#define LOGID_NONE 1		///< Used to mean 'no log'

struct log;

/** Complete format of the print function.
 * Implementation should offer simplier functions.
 */
typedef logid_t(*log_printer_func) (
	struct log *log,	///< The log object.
	enum log_level level,	///< The level under which the message must not be written.
	time_t date,	///< The date to timestamp the message with.
	logid_t *refs,	///< Other messages this one references.
	char const *format,	///< format of the message.
	...	///< Parameter for the previous format.
);

/** All functions that are overloadable.
 */
struct log_ops {
	log_printer_func print;	///< Print the message, dependant on implemtation (pure virtual, ie. undef by default, but used in ctor)
} log_ops;

/** The log structure must be defined so that you can allocate is,
 * but please don't mess directly with this. Instead, use the
 * following functions.
 * @note To log, use log->ops->print()
 */
struct log {
	struct log_ops const *ops;	///< Pointer to overloadable functions.
	enum log_level level;	///< Global log level above which messages are dropped.
};

/** Constructor.
 * @note Default log level is LOG_WARNING.
 */
int log_ctor(struct log *log);

/** Destructor.
 */
void log_dtor(struct log *log);

/** Set the current log level.
 */
static inline void log_set_level(struct log *log, enum log_level level)
{
	log->level = level;
}

/** Get the current log level.
 */
static inline enum log_level log_get_level(struct log const *log)
{
	return log->level;
}

/** Tells weither a logid is legitimate or not.
 * @note usefull in various places, internaly.
 */
static inline bool logid_reserved(logid_t id)
{
	return id == LOGID_NONE || id == LOGID_NUL;
}

/** Show a log_level as a displayable string.
 */
char const *log_level_str(enum log_level level);

#endif
