/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
/** @file
 * @brief MMaped context allocator
 * @author info@evistel.com
 * @version $Id: pool.h,v 1.1.1.1 2006/09/06 16:46:49 ccellier Exp $
 * 
 * This little library (libpool) is a replacement for usual
 * context pool  as found in many Evistel apps.
 *
 * The main known advantages upon custom per-project pool are :
 *
 * - No need to code it yourself ;
 * - Exists in object or shared object library ;
 * - The contexts are mmaped onto a file, and can be retrieved after
 *   the application crash (or exit cleanly) ;
 * - The pool size can be expended if necessary ;
 *
 * The mmaping is of course the main feature. For this to work, you must avoid to 
 * stores pointers into your contexts. pool_alloc provides you with an id that 
 * can be used to retrieve the corresponding pool with pool_get later on. Use 
 * this as a key, instead of the context address. If you need an ever growing id,
 * use pool_key2id and/or pool_key.
 *
 * Once crashed, you can copy the mmaped contexts and dump all messages. You also 
 * can restart your app with the former contexts, which is smart. If you don't 
 * want to keep contexts between crash, call pool_reset after pool_begin.  
 * Anyway, the pool contexts are checked in pool_begin, and if an error is 
 * detected (such as the free list being corrupted), all the contexts are 
 * reseted. Also, you can resize your context pool from one run to another : 
 * pool_begin take care of resizing the underlying file.
 */

#ifndef POOL_H_060822
#define POOL_H_060822

#include <inttypes.h>
#include <time.h>
#include <threadlog/commons.h>

struct pool;

/// Trace function called on various occasions
typedef void (*pool_log_func)(int prio, char const *fmt, ...);

/** Initialize the pool service.
 * Must be called first, or after a pool_dtor().
 * Any pre-existing pool_file will be checked for integrity.
 * @see pool_dtor.
 * @note To increase a pool, call pool_dtor and then call pool_begin again with
 *       a larger nb_obj. Underlying file will be expended and former objects preserved.
 * @return 0 on success, -1 on error.
 */
int pool_ctor(
	struct pool *pool,	///< The pool object.
	char const *pool_file,	///< Name of the file to use for holding contexts.
	unsigned nb_obj,	///< Number of objects that can hold a pool.
	size_t objsize,	///< Size in bytes of each object.
	pool_log_func log_cb ///< Log function called on various occasions (set to NULL to use syslog)
);

/** Closes the pool.
 * All remaining objects are still on the mmaped file.
 */
void pool_dtor(
	struct pool *pool	///< The pool object.
);

/** Empty the pool.
 * Usefull after a pool_begin if you don't care keeping old objects.
 */
void pool_reset(
	struct pool *pool	///< The pool object.
);

/** Get a slot from the pool.
 * Also, return an id that uniquely identify this slot.
 * @note Use this id instead of object's address as a key.
 * @return address of an object inside the pool.
 */
void *pool_alloc(
	struct pool *pool,	///< The pool object.
	unsigned *id	///< If not NULL, will be set to this slot's id.
);

/** Return a slot to the free list.
 */
void pool_free(
	struct pool *pool,	///< The pool object.
	void *obj	///< Points to the object inside the pool, as returned by pool_alloc.
);

/** Get a slot from its id.
 * Id must have been returned by pool_alloc.
 * @note This is a O(1) operation.
 * @see pool_alloc
 */
void *pool_get(
	struct pool *pool,	///< The pool object.
	unsigned id	///< As returned by a former pool_alloc().
);

/** Call the given function for each object in pool.
 * @param pool The pool object.
 * @param data a pointer passed to callback.
 * @param cb The callback function that will receive object address and id as
 *           arguments.
 * @return Once a callbackes do not return 0, the loop returns with that status.
 */
int pool_for_each(
	struct pool *pool,
	int (*cb)(struct pool const *, void *, unsigned, void *),
	void *data
);

/** Get the age of an object.
 * Usefull to timeout contexts.
 * @return the age (in seconds) of the object.
 */
time_t pool_age(
	struct pool *pool,	///< The pool object.
	void *obj	///< Points to the object inside the pool, as returned by pool_alloc.
);

/** Reset the age of an object.
 * Usefull to 'touch' an object.
 */
void pool_touch(
	struct pool *pool,	///< The pool object.
	void *obj	///< Points to the object inside the pool, as returned by pool_alloc.
);

/** Dump pool content on stderr.
 * @note no usefull purpose...
 */
void pool_dump(struct pool *pool);

/** Get the key from an obj.
 */
uint64_t pool_key(struct pool const *pool, void *obj);

/** Get the id within a key.
 */
static inline unsigned pool_key2id(uint64_t key)
{
	return (uint32_t) key;
}

/** This is supposed to be opaque.
 * Use pool_ctor/dtor().
 */
struct pool {
	unsigned nb_objs;		///< Number of objects in the pool, _including_ entries for internal use.
	size_t storesize;		///< Size of object reserved (greater than objsize for aligment).
	size_t objsize;		///< Requested object size.
	char *objs;			///< Pointer to the malloced memory area.
	int fd;			///< File descriptor of the mmaped file.
	size_t objs_size;		///< Size of the malloced area.
	pool_log_func log_cb; ///< Log function
};

#endif
