/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include "threadlogd.h"		// to steal DEFAULT_LOG_DIR
#include "threadlog/logdlib.h"
#include "delog.h"

/*
 * Data Definitions
 */

static char const Name[] = "$$", Id[] = "$Id$";
unsigned max_depth = UINT_MAX;
enum log_level min_level = log_DEBUG;
char const *log_dir = DEFAULT_LOG_DIR;

/*
 * Get Msg List
 */

static unsigned gml_max_depth;
static int (*gml_cb)(logid_t, unsigned, void *);
static void *gml_data;

static int foreach_referers(logid_t id, unsigned depth)
{
	if (depth > gml_max_depth) return 0;
	struct bl_curs curs;
	if (0 != bl_curs_ctor(&curs, id, log_dir)) {
		fprintf(stderr, "Cannot init backlinks cursor for %u\n", (unsigned) id);
		return -1;
	}
	int ret = 0;
	while (! logid_reserved(id = bl_curs_next(&curs))) {
		ret = gml_cb(id, depth, gml_data);
		if (ret != 0) break;
		ret = foreach_referers(id, depth + 1);
		if (ret != 0) break;
	}
	bl_curs_dtor(&curs);
	return ret;
}

static int foreach_refered(logid_t id, unsigned depth)
{
	if (depth > gml_max_depth) return 0;
	struct log_msg msg;
	if (0 != log_msg_ctor(&msg, id, log_dir)) {
		fprintf(stderr, "Cannot read message %u\n", (unsigned) id);
		return -1;
	}
	int ret = 0;
	for (unsigned r = 0; r < sizeof_array(msg.record.refs); r++) {
		if (LOGID_NUL == msg.record.refs[r]) break;
		ret = gml_cb(msg.record.refs[r], depth, gml_data);
		if (ret != 0) break;
		ret = foreach_refered(msg.record.refs[r], depth + 1);
		if (ret != 0) break;
	}
	log_msg_dtor(&msg);
	return ret;
}

static int foreach_chrono(logid_t id, unsigned depth)
{
	if (depth > gml_max_depth) return 0;
	struct chrono_curs curs;
	if (0 != chrono_curs_ctor(&curs, id, log_dir)) {
		fprintf(stderr, "Cannot init chrono cursor for %u\n", (unsigned) id);
		return -1;
	}
	int ret = 0;
	while (!logid_reserved(id = chrono_curs_next(&curs))) {
		ret = gml_cb(id, depth, gml_data);
		if (ret != 0) break;
	}
	chrono_curs_dtor(&curs);
	return ret;
}

int foreach_msg(logid_t root, enum display_mode dmode, unsigned max_depth, int (*cb)(logid_t, unsigned, void *), void *data)
{
	gml_max_depth = max_depth;
	gml_cb = cb;
	gml_data = data;
	switch (dmode) {
		case REFERERS:
			return foreach_referers(root, 0);
		case REFERED:
			return foreach_refered(root, 0);
		case CHRONO:
			return foreach_chrono(root, 0);
		default:
			return -1;
	}
}

/*
 * Command Line Interface
 */

static int print_msg(logid_t id, unsigned depth, void *data)
{
	(void) data;
	struct log_msg msg;
	if (0 != log_msg_ctor(&msg, id, log_dir)) {
		fprintf(stderr, "Cannot read message %u\n", (unsigned) id);
		return -1;
	}
	for (unsigned i = 0; i < depth; i++) {
		printf("  ");
	}
	struct tm *tm = localtime(&msg.record.date);
	if (msg.record.level <= min_level) {
		printf(
			"%02d-%02d-%02d %02d:%02d:%02d %s: %s (#%u, %u.)\n",
			tm->tm_year + 1900 - 2000, tm->tm_mon + 1, tm->tm_mday,
			tm->tm_hour, tm->tm_min, tm->tm_sec,
			log_level_str(msg.record.level), msg.string,
			(unsigned)id, msg.nb_bl);
	}
	return 0;
}

static int cli_display(logid_t id, enum display_mode disp_mode)
{
	return foreach_msg(id, disp_mode, max_depth, print_msg, NULL);
}

/*
 * Curse Interface
 */

static int interactive_display(logid_t id, enum display_mode disp_mode)
{
	// these defines the state of the display
	static logid_t history[64];	// history max depth is fixed
	logid_t cursor_pref = LOGID_NONE;
	history[0] = id;
	history[1] = LOGID_NUL;
	unsigned current = 0;
	bool unfold_all = false;
	do {
		nc_display_status_line(disp_mode, unfold_all);
		nc_display_history(history);
		logid_t selected;
		switch (nc_browse_msg(history[current], cursor_pref, disp_mode, unfold_all, &selected)) {
			case ERROR:
				return -1;
			case NOP:
				break;
			case QUIT:
				return 0;
			case SELECT:
				if (current >= sizeof(history) - 2) {
					memmove(history, history + 1, sizeof(history) - sizeof(*history));
					current--;
				}
				current++;
				history[current] = selected;
				history[current + 1] = LOGID_NUL;
				cursor_pref = LOGID_NONE;
				break;
			case CHG_DISPMODE:
				disp_mode++;
				if (disp_mode >= NB_DISP_MODES) {
					disp_mode = 0;
				}
				cursor_pref = LOGID_NONE;
				break;
			case CHG_FOLD:
				unfold_all = !unfold_all;
				cursor_pref = history[current];
				break;
			case BACK:
				if (current > 0) {
					cursor_pref = history[current];
					history[current] = LOGID_NUL;
					current--;
				}
				break;
			case LEVEL_UP:
				if (min_level > log_ALWAYS)
					min_level--;
				break;
			case LEVEL_DOWN:
				if (min_level < log_DEBUG)
					min_level++;
				break;
		}
	} while (1);
	return 0;
}

/*
 * Command Line
 */

static void syntax(void)
{
	printf(
		"delog [options]\n\n"
		"options :\n"
		"  -d dir, --dir=dir      : log directory (default: " DEFAULT_LOG_DIR ")\n"
		"  -r id, --root=id       : start viewing at this msg (default: first)\n"
		"  --mode=n|b|c           : n=normal, b=see the tree upside down, c=chronologic\n"
		"  -m n, --max-depth=n    : max depth (default: all)\n"
		"  -l n, --min-level=n    : minimum log level (default: all)\n"
		"  --cli                  : don't use ncurse\n"
		"  -h, --help             : display this help\n"
		"  -V, --version          : display version info and exit\n"
		"\n");
}

static void invalid_parameter(const char *arg)
{
	fprintf(stderr, "Invalid parameter : '%s'\n", arg);
	syntax();
	exit(EXIT_FAILURE);
}

int main(int nb_args, char **args)
{
	int ret = EXIT_FAILURE;
	logid_t root_id = LOGID_NONE;
	char *end;
	bool cli = false;
	enum display_mode disp_mode = REFERERS;
	while (1) {
		static struct option long_options[] = {
			{ "dir", required_argument, NULL, 'd' },
			{ "root", required_argument, NULL, 'r' },
			{ "mode", required_argument, NULL, 'M' },
			{ "max-depth", required_argument, NULL, 'm' },
			{ "min-level", required_argument, NULL, 'l' },
			{ "cli", no_argument, NULL, 'C' },
			{ "help", no_argument, NULL, 'h' },
			{ "version", no_argument, NULL, 'V' },
			{ 0, 0, 0, 0 },
		};
		int c = getopt_long(nb_args, args, "d:r:m:l:hV", long_options, NULL);
		switch (c) {
			case -1:
				goto end_opts;
			case 'd':
				log_dir = optarg;
				break;
			case 'r':
				root_id = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'M':
				switch (*optarg) {
					case 'n':
					case 'N':
						disp_mode = REFERERS;
						break;
					case 'b':
					case 'B':
						disp_mode = REFERED;
						break;
					case 'c':
					case 'C':
					case 't':
					case 'T':
						disp_mode = CHRONO;
						break;
					default:
						invalid_parameter(optarg);
						break;
				}
				break;
			case 'm':
				max_depth = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'l':
				min_level = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'C':
				cli = true;
				break;
			case 'h':
				syntax();
				return EXIT_SUCCESS;
			case 'V':
				printf("Loger " TOSTR(VERSION) "\n%s, version %s\n", Name, Id);
				return EXIT_SUCCESS;
			case ':':
				fprintf(stderr, "Missing parameter\n");
				syntax();
				exit(EXIT_FAILURE);
			case '?':
				fprintf(stderr, "Bad syntax\n");
				syntax();
				exit(EXIT_FAILURE);
			default:
				fprintf(stderr, "?? getopt returned character code %d ??\n", c);
				exit(EXIT_FAILURE);
		}
	}
end_opts:
	// find root
	if (root_id == LOGID_NONE) {
		root_id = path_first_id(log_dir);
	}
	// try ncurses init
	if (cli || 0 != nc_begin(NULL)) {
		ret = 0 == cli_display(root_id, disp_mode) ? EXIT_SUCCESS : EXIT_FAILURE;
	} else {
		ret = 0 == interactive_display(root_id, disp_mode) ? EXIT_SUCCESS : EXIT_FAILURE;
		nc_end();
	}
	return ret;
}
