/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <assert.h>
#include <dlfcn.h>
#include <curses.h>
#include "threadlog/commons.h"
#include "threadlog/logdlib.h"
#include "delog.h"

/*
 * Begin / End ncurses
 */

#define DLSYM(x) static __typeof__(x) *dl_##x
DLSYM(initscr);
DLSYM(endwin);
DLSYM(has_colors);
DLSYM(start_color);
DLSYM(keypad);
DLSYM(nonl);
DLSYM(cbreak);
DLSYM(echo);
DLSYM(init_pair);
DLSYM(getch);
DLSYM(attrset);
DLSYM(wattrset);
DLSYM(stdscr);
DLSYM(mvprintw);
DLSYM(printw);
DLSYM(move);
DLSYM(refresh);
DLSYM(COLS);
DLSYM(LINES);
DLSYM(attron);
DLSYM(attroff);
DLSYM(beep);
DLSYM(erase);
DLSYM(hline);
DLSYM(acs_map);
DLSYM(derwin);
DLSYM(delwin);
DLSYM(box);
DLSYM(mvwprintw);
DLSYM(chgat);
DLSYM(curs_set);
DLSYM(clrtobot);

static void *dl_handle;

enum {
	STATUS_COLOR = 1,
	HISTORY_COLOR,
	ERROR_COLOR,
	WARNING_COLOR,
	NOTICE_COLOR,
	HELP_COLOR,
};

static void *Dlsym(char const *name)
{
	char *error;
	dlerror();
	void *sym = dlsym(dl_handle, name);
	if (NULL != (error = dlerror())) {
		fprintf(stderr, "%s\n", error);
		exit(EXIT_FAILURE);
	}
	return sym;
}

int nc_begin(char const *lib)
{
#	define DLLOADSYM(x) dl_##x = Dlsym( TOSTR(x) )
	dl_handle = dlopen(lib ? lib : "libcurses.so", RTLD_LAZY);
	if (! dl_handle) {
		fprintf(stderr, "%s\n", dlerror());
		return -1;
	}

	DLLOADSYM(initscr);
	DLLOADSYM(endwin);
	DLLOADSYM(has_colors);
	DLLOADSYM(start_color);
	DLLOADSYM(keypad);
	DLLOADSYM(nonl);
	DLLOADSYM(cbreak);
	DLLOADSYM(echo);
	DLLOADSYM(init_pair);
	DLLOADSYM(getch);
	DLLOADSYM(attrset);
	DLLOADSYM(wattrset);
	DLLOADSYM(stdscr);
	DLLOADSYM(mvprintw);
	DLLOADSYM(printw);
	DLLOADSYM(move);
	DLLOADSYM(refresh);
	DLLOADSYM(COLS);
	DLLOADSYM(LINES);
	DLLOADSYM(attron);
	DLLOADSYM(attroff);
	DLLOADSYM(beep);
	DLLOADSYM(erase);
	DLLOADSYM(hline);
	DLLOADSYM(acs_map);
	DLLOADSYM(derwin);
	DLLOADSYM(delwin);
	DLLOADSYM(box);
	DLLOADSYM(mvwprintw);
	DLLOADSYM(chgat);
	DLLOADSYM(curs_set);
	DLLOADSYM(clrtobot);

	(void)dl_initscr();
	dl_keypad(*dl_stdscr, TRUE);
	(void)dl_nonl();
	(void)dl_cbreak;
	(void)dl_echo();
	(void)dl_curs_set(0);
	if (dl_has_colors()) {
		dl_start_color();
		dl_init_pair(STATUS_COLOR, COLOR_WHITE, COLOR_BLUE);
		dl_init_pair(HISTORY_COLOR, COLOR_YELLOW, COLOR_BLACK);
		dl_init_pair(ERROR_COLOR, COLOR_RED, COLOR_BLACK);
		dl_init_pair(WARNING_COLOR, COLOR_YELLOW, COLOR_BLACK);
		dl_init_pair(NOTICE_COLOR, COLOR_GREEN, COLOR_BLACK);
		dl_init_pair(HELP_COLOR, COLOR_YELLOW, COLOR_BLUE);
	}
	return 0;
}

void nc_end(void)
{
	dl_endwin();
	dlclose(dl_handle);
}

/*
 * User Interface
 */

/*
 * Private Functions
 */

static void print_msg_long(logid_t id)
{
	struct log_msg msg;
	if (0 != log_msg_ctor(&msg, id, log_dir)) {
		fprintf(stderr, "Cannot read message %u\n", (unsigned)id);
		return;
	}
	struct tm *tm = localtime(&msg.record.date);
	dl_printw(
		"%02d-%02d-%02d %02d:%02d:%02d %s: %s (#%u, %u.)\n",
		tm->tm_year + 1900 - 2000, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec,
		log_level_str(msg.record.level), msg.string,
		(unsigned)msg.id, msg.nb_bl);
}

void nc_display_status_line(enum display_mode disp_mode, bool unfold_all)
{
	dl_erase();
	dl_attrset(COLOR_PAIR(STATUS_COLOR));
	time_t now = time(NULL);
	struct tm *tm = localtime(&now);

	dl_mvprintw(
		0, 0, "Delog v" TOSTR(VERSION) " - Time: %02d-%02d-%02d %02d:%02d - Mode: %s, %s - Min log level: %s - press H for Help",
		tm->tm_year + 1900 - 2000, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min,
		disp_mode == REFERED ? "Show refered" : disp_mode == REFERERS ? "Show referers" : "Chronologic",
		unfold_all ? "unfolded" : "folded",
		log_level_str(min_level));
}

void nc_display_history(logid_t *history)
{
	dl_move(1, 0);
	dl_attrset(COLOR_PAIR(HISTORY_COLOR));
	while (LOGID_NUL != *history) {
		if (LOGID_NONE != *history) {
			print_msg_long(*history);
		}
		history++;
	}
	dl_hline((*dl_acs_map)['q'], *dl_COLS);
	int x, y;
	getyx(*dl_stdscr, y, x);
	dl_move(y + 1, 0);
}

static char const *some_spaces(unsigned nb)
{
	static char const spaces[] = "                                                                                 ";
	if (nb > sizeof(spaces))
		return spaces;
	return spaces + sizeof(spaces) - 1 - nb;
}

static void print_msg_short(struct log_msg *msg,
		unsigned depth, unsigned y)
{
	static char str[1024];	// Max line size
	switch (msg->record.level) {
		case log_ERROR:
			dl_attrset(COLOR_PAIR(ERROR_COLOR));
			break;
		case log_WARNING:
			dl_attrset(COLOR_PAIR(WARNING_COLOR));
			break;
		case log_NOTICE:
			dl_attrset(COLOR_PAIR(NOTICE_COLOR));
		default:
			dl_attrset(A_NORMAL);
			break;
	}
	struct tm *tm = localtime(&msg->record.date);
	snprintf(
		str, sizeof(str),
		"%02d-%02d-%02d %02d:%02d:%02d %s (#%u, %u.): %s",
		tm->tm_year + 1900 - 2000, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec, some_spaces(depth),
		(unsigned)msg->id, msg->nb_bl, msg->string);
	int d;
	for (d = 0; d < *dl_COLS && d < (int)sizeof(str) - 1; d++) {
		if (str[d] == '\n') str[d] = '~';
		if (str[d] == '\0') break;
	}
	if (d == *dl_COLS && str[d] != '\0') {
		str[d] = '\0';
		if (*dl_COLS >= 3) {	// just to be sure
			str[d - 1] = '.';
			str[d - 2] = '.';
			str[d - 3] = '.';
		}
	}
	dl_mvprintw(y, 0, "%s", str);
}

static void print_help(void)
{
	static char const *help_str[] = {
		"Help",
		"----",
		"",
		"Navigation in the log tree :",
		"  On top of the screen, you are shown all selected messages,",
		"  in order you selected them. This is you navigation history.",
		"  The last selected message is the root of the current tree.",
		"  The rest of the screen lists all messages referencing this",
		"  root message. You can also hit the Tab key to display the",
		"  list of the message referenced by this root message.",
		"  Use this list to select another root, or use Backspace to go",
		"  back." "",
		"Keys :",
		"  Up/Down : Move cursor",
		"  Enter/Rigth : Make the selected message root",
		"  Backspace/Left : Get back to previous root",
		"  Space : (un)fold all messages",
		"  Tab : swap between displaying root's references or referers",
		"  +/- : Change minimal log level",
		"  H : this help (you obviously already knew",
		"  Q/F10 : Quit",
		"",
		"press any key to close this help",
	};
#	define HELPWIN_W 65
#	define HELPWIN_H sizeof_array(help_str)+2
	WINDOW *help_win = dl_derwin(*dl_stdscr, HELPWIN_H, HELPWIN_W, (*dl_LINES - HELPWIN_H) / 2, (*dl_COLS - HELPWIN_W) / 2);
	dl_wattrset(help_win, COLOR_PAIR(HELP_COLOR));
	dl_box(help_win, 0, 0);
	for (unsigned y = 0; y < sizeof_array(help_str); y++) {
		dl_mvwprintw(help_win, y + 1, 1, "                                                               ");
		dl_mvwprintw(help_win, y + 1, 1, help_str[y]);
	}
	dl_delwin(help_win);
}

static void print_cursor(unsigned y)
{
	dl_move(y, 0);
	dl_chgat(-1, A_REVERSE, 0, NULL);
}

static void erase_cursor(unsigned y)
{
	dl_move(y, 0);
	dl_chgat(-1, 0, 0, NULL);
}

struct fem_param {
	unsigned page_h;
	logid_t pref_id;
	unsigned cursor_idx;
	bool cursor_set;
	unsigned len;
	unsigned oldest;
	struct fem_msg {
		struct log_msg msg;
		unsigned depth;
	} *msgs;
	bool bottom;		// when looking for cursor, return msgs with cursor at bottom
};

static int find_first(logid_t id, unsigned depth, void *data)
{
	struct fem_param *fp = data;
	struct log_msg *msg = &fp->msgs[fp->len].msg;
	if (0 != log_msg_ctor(msg, id, log_dir)) {
		return -1;
	}
	if (msg->record.level > min_level) {
		log_msg_dtor(msg);
		return 0;
	}
	fp->msgs[fp->len].depth = depth;
	fp->len++;
	if (fp->len >= fp->page_h) return 1;
	return 0;
}

// Beware that the same id can be encountered multiple times !
static int find_cursor(logid_t id, unsigned depth, void *data)
{
	struct fem_param *fp = data;
	struct log_msg *msg;
	struct log_msg tmp;
	// Initial tests
	if (! fp->bottom) {
		if (fp->len == 0 && id != fp->pref_id) return 0;
		msg = &fp->msgs[fp->len].msg;
	} else {
		if (fp->len < fp->page_h) {
			msg = &fp->msgs[fp->len].msg;
		} else {
			msg = &tmp;
		}
	}
	// Read and filter message
	if (0 != log_msg_ctor(msg, id, log_dir)) {
		return -1;
	}
	if (msg->record.level > min_level) {	// not interrested
		log_msg_dtor(msg);
		if (id == fp->pref_id) return 1;
		return 0;
	}
	// Keep it
	if (! fp->bottom) {
		if (id == fp->pref_id) {
			fp->cursor_idx = fp->len;
			fp->cursor_set = true;
		}
		fp->msgs[fp->len].depth = depth;
		fp->len++;
		if (fp->len >= fp->page_h) return 1;
	} else {
		unsigned idx;
		if (fp->len < fp->page_h) {
			idx = fp->len;
			fp->len ++;
		} else {
			idx = fp->oldest;
			fp->oldest = (fp->oldest + 1) % fp->page_h;
			fp->msgs[idx].msg = *msg;
		}
		fp->msgs[idx].depth = depth;
		if (id == fp->pref_id) {
			fp->cursor_idx = idx;
			fp->cursor_set = true;
		}
		if (fp->cursor_set && fp->len >= fp->page_h) return 1;
	}
	return 0;
}

enum browse_action nc_browse_msg(logid_t root, logid_t pref_id, enum display_mode disp_mode, bool unfold_all, logid_t *selected)
{
	int x_beg, y_beg;
	getyx(*dl_stdscr, y_beg, x_beg);
	dl_attrset(A_NORMAL);
	unsigned page_h = *dl_LINES - y_beg;
	unsigned my_max_depth = unfold_all ? max_depth : 0;

	// look for cursor_pref and a message to display
	struct fem_msg msgs[page_h];
	struct fem_param fem_param = {
		.page_h = page_h,
		.pref_id = pref_id,
		.msgs = msgs,
		.bottom = false,
	};
find_msg:
	fem_param.len = fem_param.oldest = 0;
	fem_param.cursor_set = false;
	unsigned cursor_line = 0;
	if (LOGID_NONE != fem_param.pref_id) {
		if (-1 == foreach_msg(root, disp_mode, my_max_depth, find_cursor, &fem_param)) {
			return ERROR;
		}
	}
	if (fem_param.len == 0) {	// pref cursor not found, will go from first_ids.
		if (-1 == foreach_msg(root, disp_mode, my_max_depth, find_first, &fem_param)) {
			return ERROR;
		}
	}
	bool has_curs = fem_param.len > 0;
	if (has_curs && !fem_param.cursor_set) {	// we have some messages to display, but no cursor yet
		fem_param.cursor_idx = 0;
	}
	if (has_curs) {
		if (fem_param.bottom) {	// adjust cursor_line
			if (fem_param.oldest > fem_param.cursor_idx) {
				cursor_line = (fem_param.cursor_idx + page_h) - fem_param.oldest;
			} else {
				cursor_line = fem_param.cursor_idx - fem_param.oldest;
			}
		} else {
			cursor_line = fem_param.cursor_idx;
		}
	}
	enum browse_action ret = NOP;
	do {
		// display
		dl_move(y_beg, x_beg);
		dl_clrtobot();
		for (unsigned i = 0; i < fem_param.len; i++) {
			unsigned idx = (i + fem_param.oldest) % page_h;
			print_msg_short(&msgs[idx].msg, msgs[idx].depth, y_beg + i);
		}
		if (has_curs) print_cursor(y_beg + cursor_line);
		dl_refresh();
		// Now wait a key. We handle cursor internaly
		int key = dl_getch();
		switch (key) {
			case KEY_DOWN:
				if (has_curs) {
					if (cursor_line < fem_param.len - 1) {
						erase_cursor(y_beg + cursor_line);
						cursor_line++;
						fem_param.cursor_idx = (fem_param.cursor_idx + 1) % page_h;
						print_cursor(y_beg + cursor_line);
					} else if (fem_param.len == page_h) {
						fem_param.pref_id = msgs[fem_param.cursor_idx].msg.id;
						fem_param.bottom = false;
						goto find_msg;
					}
				}
				break;
			case KEY_UP:
				if (has_curs) {
					if (cursor_line > 0) {
						erase_cursor(y_beg + cursor_line);
						cursor_line--;
						fem_param.cursor_idx = fem_param.cursor_idx > 0 ? fem_param.cursor_idx - 1 : page_h - 1;
						print_cursor(y_beg + cursor_line);
					} else {
						fem_param.pref_id = msgs[fem_param.cursor_idx].msg.id;
						fem_param.bottom = true;
						goto find_msg;
					}
				}
				break;
			case KEY_NPAGE:
				if (has_curs) {
					if (fem_param.len == page_h) {
						unsigned idx = (fem_param.oldest + page_h - 1) % page_h;
						fem_param.pref_id = msgs[idx].msg.id;
						fem_param.bottom = false;
						goto find_msg;
					} else {
						erase_cursor(y_beg + cursor_line);
						fem_param.cursor_idx = (fem_param.oldest - 1) % page_h;
						cursor_line = fem_param.len - 1;
						print_cursor(y_beg + cursor_line);
					}
				}
				break;
			case KEY_PPAGE:
				if (has_curs) {
					fem_param.pref_id = msgs[fem_param.oldest].msg.id;
					fem_param.bottom = true;
					goto find_msg;
				}
				break;
			case 'q':
			case 'Q':
			case KEY_F(10):
			case KEY_F(12):
				ret = QUIT;
				break;
			case KEY_RIGHT:
			case KEY_SELECT:
			case KEY_ENTER:
			case 13:
				if (has_curs) {
					*selected = msgs[fem_param.cursor_idx].msg.id;
					ret = SELECT;
				}
				break;
			case ' ':
				ret = CHG_FOLD;
				break;
			case '\t':
				ret = CHG_DISPMODE;
				break;
			case KEY_LEFT:
			case KEY_BACKSPACE:
				ret = BACK;
				break;
			case '+':
				ret = LEVEL_UP;
				break;
			case '-':
				ret = LEVEL_DOWN;
				break;
			case 'h':
			case 'H':
				print_help();
				dl_getch();
				break;
			default:
				dl_beep();
				break;
		}
	} while (ret == NOP);
	for (unsigned i = 0; i < fem_param.len; i++) {
		log_msg_dtor(&msgs[i].msg);
	}
	return ret;
}
