/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include "threadlog/logc.h"
#include "threadlogd.h"		// steal DEFAULT_PORT

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";
struct logc logc;
#define DEFAULT_LOG_LEVEL log_WARNING
#define DEFAULT_LOCAL_ID "logcmd"

/*
 * Command line
 */

static void syntax(void)
{
	printf(
		"log [options] message\n\n"
		"options :\n"
		"  -H host, --host=host   : loger host (default: localhost)\n"
		"  -p port, --port=port   : loger port (default: " TOSTR(DEFAULT_PORT) ")\n"
		"  -l n, --level=n        : log level (%d = ERROR to %d = DEBUG, default = " TOSTR(DEFAULT_LOG_LEVEL) ")\n"
		"  --local=str            : local_id used to help identify login source (default: " DEFAULT_LOCAL_ID ")\n"
		"  -h, --help             : display this help\n"
		"  -V, --version          : display version info and exit\n"
		"\n",
		log_ALWAYS, log_DEBUG);
}

static void invalid_parameter(const char *arg)
{
	fprintf(stderr, "Invalid parameter : '%s'\n", arg);
	syntax();
	exit(EXIT_FAILURE);
}

int main(int nb_args, char **args)
{
	int ret = EXIT_FAILURE;
	char const *host = "localhost";
	unsigned short port = DEFAULT_PORT;
	enum log_level level = DEFAULT_LOG_LEVEL;
	char local_id[LOG_LOCAL_ID_MAX] = DEFAULT_LOCAL_ID;
	char *end;
	while (1) {
		static struct option long_options[] = {
			{ "host", required_argument, NULL, 'H' },
			{ "port", required_argument, NULL, 'p' },
			{ "level", required_argument, NULL, 'l' },
			{ "local", required_argument, NULL, 'L' },
			{ "help", no_argument, NULL, 'h' },
			{ "version", no_argument, NULL, 'V' },
			{ 0, 0, 0, 0 },
		};
		int c = getopt_long(nb_args, args, "H:p:l:L:hV", long_options, NULL);
		switch (c) {
			case -1:
				goto end_opts;
			case 'H':
				host = optarg;
				break;
			case 'p':
				port = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'l':
				level = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'L':
				strncpy(local_id, optarg, sizeof(local_id));
				break;
			case 'h':
				syntax();
				return EXIT_SUCCESS;
			case 'V':
				printf("Loger " TOSTR(VERSION) "\n%s, version %s\n", Name, Id);
				return EXIT_SUCCESS;
			case ':':
				fprintf(stderr, "Missing parameter\n");
				syntax();
				exit(EXIT_FAILURE);
			case '?':
				fprintf(stderr, "Bad syntax\n");
				syntax();
				exit(EXIT_FAILURE);
			default:
				fprintf(stderr, "?? getopt returned character code %d ??\n", c);
				exit(EXIT_FAILURE);
		}
	}
end_opts:
	// there must be some parameters left for message
	if (optind >= nb_args) {
		fprintf(stderr, "No message ?\n");
		syntax();
		exit(EXIT_FAILURE);
	}
	// inits
	struct logc logc;
	if (0 != logc_ctor(&logc, &local_id, host, port)) {
		fprintf(stderr, "Cannot init logc\n");
		goto err_0;
	}
	char str[LOG_STRING_MAX] = "";
	bool space = false;
	while (optind < nb_args) {
		if (space) strncat(str, " ", sizeof(str) - 1);
		space = true;
		strncat(str, args[optind++], sizeof(str) - 1);
	}
	// Beware that str may contain formating chars. Better use '%s'.
	log_set_level(&logc.log, level);
	if (LOGID_NONE ==
			logc.log.ops->print(&logc.log, level, time(NULL), NULL, "%s", str)) {
		fprintf(stderr, "Cannot print log\n");
		goto err_9;
	}
	// quit
	ret = EXIT_SUCCESS;
err_9:
	logc_dtor(&logc);
err_0:
	return ret;
}
