/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "threadlogd.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";
static int sockfd;

struct search_client_param {
	struct client *client;
	struct sockaddr_in const *addr;
};

/*
 * Private Functions
 */

static int init_socket(unsigned short port)
{
	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == sockfd) {
		MY_ERROR("Cannot create socket : %s", strerror(errno));
		return -1;
	}
	struct sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(port);
	if (0 != bind(sockfd, (struct sockaddr const *)&addr, sizeof(addr))) {
		MY_ERROR("Cannot bind port : %s", strerror(errno));
		close(sockfd);
		return -1;
	}
	return 0;
}

static int close_socket(void)
{
	// TODO: do something about pending clients ?
	return close(sockfd);
}

static int search_client(struct pool const *pool, void *obj, unsigned id, void *data)
{
	(void) pool;
	(void) id;
	struct client *client = obj;
	struct search_client_param *param = data;
	if (
		client->address.sin_port == param->addr->sin_port &&
		client->address.sin_addr.s_addr == param->addr->sin_addr.s_addr
	) {
		param->client = client;
		return 1;
	}
	return 0;
}

static struct client *find_client(struct sockaddr_in const *from)
{
	struct search_client_param param = {
		.client = NULL,
		.addr = from,
	};
	if (1 == pool_for_each(&pool, search_client, &param)) {
		return param.client;
	}
	return NULL;
}

static void map_seq_to_logids(logid_t *refs, logid_t *seq_refs, struct client *client)
{
	unsigned i = 0;
	unsigned seqi = 0;
	while (seqi++ < LOG_REFS_MAX) {
		if (seq_refs[seqi] == LOGID_NUL) break;
		if (seq_refs[seqi] == LOGID_NONE) continue;
		struct seq2id *seq2id = client_find_seq2id(client, seq_refs[seqi]);
		if (seq2id) {
			refs[i++] = seq2id->id;
			client_promote_seq2id(client, seq2id);
		}
	}
	if (i < LOG_REFS_MAX) {
		refs[i] = LOGID_NUL;
	}
}

static int process_dtg(struct dtg_max *dtg, struct sockaddr_in const *from)
{
	struct client *client = find_client(from);
	if (! client) {
		client = pool_alloc(&pool, NULL);
		if (! client) {		// TODO: timeout old clients
			MY_ERROR("No more free clients");
			return -1;
		}
		if (0 != client_ctor(client, from, dtg->hdr.local_id)) {
			pool_free(&pool, client);
			return -1;
		}
		client->connect_id =
			logd.log.ops->print(&logd.log, log_ALWAYS, time(NULL), NULL, "New client : %s", client_name(client));
	} else {
		if (0 != strncmp(dtg->hdr.local_id, client->local_id, sizeof(dtg->hdr.local_id))) {
			MY_WARNING("Wrong local_id, changing it from %.*s to %.*s",
				sizeof(client->local_id), client->local_id,
				sizeof(dtg->hdr.local_id), dtg->hdr.local_id);
			client_reset(client, dtg->hdr.local_id);
		}
	}
	struct seq2id *seq2id = client_find_seq2id(client, dtg->hdr.this_logid);
	if (seq2id) {
		MY_ERROR("SeqId %u received twice for client %s",
			dtg->hdr.this_logid, client_name(client));
		return -1;
	}
	logid_t refs[LOG_REFS_MAX];
	map_seq_to_logids(refs, dtg->hdr.refs, client);	// will strip LOGID_NONEs, and will promote referenced seqids
	if (refs[0] == LOGID_NUL) {	// by default, attach to client connect_id
		refs[0] = client->connect_id;
		refs[1] = LOGID_NUL;
	}
	dtg->hdr.str[dtg->hdr.str_size] = '\0';
	logid_t logid =
		logd.log.ops->print(&logd.log, dtg->hdr.level, dtg->hdr.date, refs, dtg->hdr.str);
	if (logid == LOGID_NONE) {
		return -1;
	}
	client_add_seq2id(client, dtg->hdr.this_logid, logid);
	return 0;
}

/*
 * Public Functions
 */

int service(unsigned short port)
{
	if (0 != init_socket(port)) {
		return -1;
	}
	while (!terminate) {
		struct sockaddr from;
		struct dtg_max dtg;
		socklen_t fromlen = sizeof(from);
		ssize_t rs = recvfrom(sockfd, &dtg, sizeof(dtg), 0, &from, &fromlen);
		if (rs < 0) {
			if (errno != EINTR) {
				MY_ERROR("Error while receiving : %s", strerror(errno));
			}
		} else if ((size_t) rs < sizeof(dtg.hdr)) {
			MY_WARNING("Short read : only %zd bytes", rs);
		} else if (fromlen != sizeof(struct sockaddr_in) || from.sa_family != AF_INET) {
			MY_WARNING("Bad family peer address : %d", (int) from.sa_family);
		} else {
			(void) process_dtg(&dtg, (struct sockaddr_in *) &from);
		}
	}
	return close_socket();
}
