/* This file is part of Threadloger.
 *
 * Threadloger is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Threadloger is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Threadloger; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <syslog.h>
#include "threadlogd.h"
#include "threadlog/pool.h"
#include "threadlog/logdlib.h"

/*
 * Data Definitions
 */

static char const Name[] = "$Name$", Id[] = "$Id$";
volatile sig_atomic_t terminate = 0;
struct logd logd;
struct pool pool;

/*
 * Command line
 */

static void syntax(void)
{
	printf("threadlogd [options]\n\n"
		"options :\n"
		"  -d dir, --dir=dir      : log directory (default: " DEFAULT_LOG_DIR ")\n"
		"  -s n, --period=n       : rotation period in secs (default: " TOSTR(DEFAULT_LOG_PERIOD) ")\n"
		"  --pool=file            : pool mmaped file (default: " DEFAULT_POOL_FILE ")\n"
		"  -p n, --port=n         : service port number (default: " TOSTR(DEFAULT_PORT) ")\n"
		"  --level=n              : minimal log level (default: 2=WARNING)\n"
		"  -L, --max-logfile=size : max size, in Mb, of logfile (default: " TOSTR(DEFAULT_LOG_SIZE) ")\n"
		"  -h, --help             : display this help\n"
		"  -V, --version          : display version info and exit\n"
		"\n" );
}

static void invalid_parameter(const char *arg)
{
	fprintf(stderr, "Invalid parameter : '%s'\n", arg);
	syntax();
	exit(EXIT_FAILURE);
}

static void sighandler(int signum)
{
	(void) signum;
	terminate = 1;
}

static void my_pool_log(int prio, char const *fmt, ...)
{
	// TODO: use the GNU extention to build function calls instead of using vsnprintf
	static char str[1024];
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(str, sizeof(str), fmt, ap);
	MY_LOG(log_prio2level(prio), str);
	va_end(ap);
}

int main(int nb_args, char **args)
{
	int ret = EXIT_FAILURE;
	char const *log_dir = DEFAULT_LOG_DIR;
	char const *pool_file = DEFAULT_POOL_FILE;
	unsigned log_period = DEFAULT_LOG_PERIOD;
	unsigned short port = DEFAULT_PORT;
	unsigned long max_log = DEFAULT_LOG_SIZE;
	enum log_level level = log_WARNING;
	char *end;
	while (1) {
		static struct option long_options[] = {
			{ "dir", required_argument, NULL, 'd' },
			{ "period", required_argument, NULL, 's' },
			{ "pool", required_argument, NULL, 'P' },
			{ "port", required_argument, NULL, 'p' },
			{ "level", required_argument, NULL, 'e' },
			{ "max-logfile", required_argument, NULL, 'L' },
			{ "help", no_argument, NULL, 'h' },
			{ "version", no_argument, NULL, 'V' },
			{ 0, 0, 0, 0 },
		};
		int c = getopt_long(nb_args, args, "d:s:P:p:L:hV", long_options, NULL);
		switch (c) {
			case -1:
				goto end_opts;
			case 'd':
				log_dir = optarg;
				break;
			case 's':
				log_period = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'P':
				pool_file = optarg;
				break;
			case 'p':
				port = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'e':
				level = strtoul(optarg, &end, 0);
				if (*end != '\0') invalid_parameter(optarg);
				break;
			case 'L':
				if (!sscanf(optarg, "%ld", &max_log)) {
					invalid_parameter(optarg);
				}
				break;
			case 'h':
				syntax();
				return EXIT_SUCCESS;
			case 'V':
				printf("ThreadLogd " TOSTR(VERSION) "\n%s, version %s\n", Name, Id);
				return EXIT_SUCCESS;
			case ':':
				fprintf(stderr, "Missing parameter\n");
				syntax();
				exit(EXIT_FAILURE);
			case '?':
				fprintf(stderr, "Bad syntax\n");
				syntax();
				exit(EXIT_FAILURE);
			default:
				fprintf(stderr, "?? getopt returned character code %d ??\n", c);
				exit(EXIT_FAILURE);
		}
	}
end_opts:;
	// signal handler
	struct sigaction action = {
		.sa_handler = sighandler,
		.sa_flags = 0,
	};
	sigemptyset(&action.sa_mask);
	if (-1 == sigaction(SIGTERM, &action, NULL)) {
		fprintf(stderr, "Cannot set signal handler for TERM : %s\n", strerror(errno));
		goto err_0;
	}
	if (-1 == sigaction(SIGINT, &action, NULL)) {
		fprintf(stderr, "Cannot set signal handler for INT : %s\n", strerror(errno));
		goto err_0;
	}
	action.sa_handler = SIG_IGN;
	if (-1 == sigaction(SIGPIPE, &action, NULL)) {
		fprintf(stderr, "Cannot ignore SIGPIPE : %s\n", strerror(errno));
		goto err_0;
	}
	// inits
	if (0 != logd_ctor(&logd, log_dir, log_period)) {
		fprintf(stderr, "Cannot init logd\n");
		goto err_0;
	}
	log_set_level(&logd.log, level);
	if (0 != pool_ctor(&pool, pool_file, 128, sizeof(struct client), my_pool_log)) {
		fprintf(stderr, "Cannot init pool\n");
		goto err_1;
	}
	// may write some output to syslog
	openlog("threadlogd", LOG_PID, LOG_DAEMON);
	// launch service
	if (0 != service(port))
		goto err_9;
	// quit
	ret = EXIT_SUCCESS;
err_9:
	pool_dtor(&pool);
err_1:
	logd_dtor(&logd);
err_0:
	closelog();
	return ret;
}
